/*
 * GroupLens Common Utilities
 * Copyright © 2011 Regents of the University of Minnesota
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *  * Neither the name of the University of Minnesota nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software has been partly funded by NSF grant IIS 08-08692.
 */
package org.grouplens.common.dto;

import java.io.Reader;
import java.io.Writer;

/**
 * General interface to support the input and output of Dto's stored within a
 * DtoContainer.
 */
public interface DtoContentHandler {
    /**
     * Serialize the Dtos within <tt>dto</tt> to a String. If dto contains no
     * Dtos the response is implementation specific.
     * 
     * @param dto
     * @return The DtoContainer in serialized form
     * @throws NullPointerException if result is null
     */
    public <T extends Dto> String toString(DtoContainer<T> dto);

    /**
     * As {@link #toString(DtoContainer)} but the serialized content is sent to
     * the given Writer.
     * 
     * @param dto
     * @param writer
     */
    public <T extends Dto> void toString(DtoContainer<T> dto, Writer writer);

    /**
     * <p>
     * Deserialize the given content and store the resultant Dtos into
     * <tt>result</tt>. It can be assumed that <tt>content</tt> is in an
     * appropriate form, as if it were generated by
     * {@link #toString(DtoContainer)} by this content handler.
     * </p>
     * <p>
     * The following general contract should be upheld by the content handler:
     * <ul>
     * <li>An unexpected root fails with an exception</li>
     * <li>A property stored in content, but not defined in the expected Dto
     * type should be ignored. If a property is defined in the Dto, but is
     * excluded from import it should also be ignored</li>
     * <li>A property that has a value with an inappropriate data type fails
     * with an exception</li>
     * <li>An expected root that is otherwise empty is an instantiated, but
     * blank, Dto of correct type</li>
     * <li>Empty/blank input is treated as a null Dto</li>
     * <li>Properties in the content that have a blank or empty value have a
     * reasonable default value.</li>
     * <ul>
     * <li>For strings: the empty string</li>
     * <li>Numeric types: 0</li>
     * <li>Character: '\0'</li>
     * <li>Boolean: false</li>
     * <li>Dto: instantiated Dto of correct type, with no properties set</li>
     * <li>Array: 0-length array of appropriate element type</li>
     * </ul>
     * </ul>
     * </p>
     * <p>
     * If the content handler supports null values represented in the serialized
     * content, these should be preserved in Java after being de-serialized by
     * this method.
     * 
     * @param content
     * @param result
     * @throws ParseException if content cannot be parsed or is otherwise
     *             invalid
     */
    public <T extends Dto> void fromString(String content, DtoContainer<T> result);

    /**
     * As {@link #fromString(String, DtoContainer)} but with a Reader instead.
     * @param reader
     * @param result
     * @throws ParseException if content cannot be parsed or is otherwise
     *             invalid
     */
    public <T extends Dto> void fromString(Reader content, DtoContainer<T> container);
}
