/*
 * GroupLens Common Utilities
 * Copyright © 2011 Regents of the University of Minnesota
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *  * Neither the name of the University of Minnesota nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software has been partly funded by NSF grant IIS 08-08692.
 */
package org.grouplens.common.dto;

import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.grouplens.common.dto.Dto.Exclude;
import org.grouplens.common.dto.Dto.ExcludeType;
import org.grouplens.common.dto.Dto.Named;

import com.google.gson.ExclusionStrategy;
import com.google.gson.FieldAttributes;
import com.google.gson.FieldNamingStrategy;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonDeserializationContext;
import com.google.gson.JsonDeserializer;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParseException;
import com.google.gson.JsonPrimitive;
import com.google.gson.JsonSerializationContext;
import com.google.gson.JsonSerializer;

/**
 * A DtoContentHandler that uses the Gson library to support the JSON content
 * type.
 */
public class JsonDtoContentHandler implements DtoContentHandler {
    private static final Object VOID = new Object();
    
    private final Gson exporter;
    private final Gson importer;
    private final boolean ignoreExclude;
    
    private final DtoNamingStrategy modifier;
    private final ConcurrentMap<Class<? extends Dto>, Object> registeredTypes;
    private final DtoContainerSerializer serializer;

    
    public JsonDtoContentHandler() {
        this(false);
    }
    
    public JsonDtoContentHandler(boolean prettyPrint) {
        this(prettyPrint, false);
    }
    
    public JsonDtoContentHandler(boolean prettyPrint, boolean ignoreExclude) {
        this.ignoreExclude = ignoreExclude;
        
        modifier = new DtoNamingStrategy();
        registeredTypes = new ConcurrentHashMap<Class<? extends Dto>, Object>();
        serializer = new DtoContainerSerializer();

        GsonBuilder exportBuilder = new GsonBuilder().disableHtmlEscaping()
                                                     .setExclusionStrategies(new DtoExclusionStrategy(false))
                                                     .setFieldNamingStrategy(modifier)
                                                     .registerTypeAdapter(DtoContainer.class, serializer);
        if (prettyPrint)
            exportBuilder.setPrettyPrinting();
        exporter = exportBuilder.create();
        
        importer = new GsonBuilder().disableHtmlEscaping()
                                    .setExclusionStrategies(new DtoExclusionStrategy(true))
                                    .setFieldNamingStrategy(modifier)
                                    .registerTypeAdapter(DtoContainer.class, serializer)
                                    .registerTypeAdapter(Character.class, new CharacterDeserializer())
                                    .registerTypeAdapter(Short.class, new NumberDeserializer())
                                    .registerTypeAdapter(Integer.class, new NumberDeserializer())
                                    .registerTypeAdapter(Long.class, new NumberDeserializer())
                                    .registerTypeAdapter(Byte.class, new NumberDeserializer())
                                    .registerTypeAdapter(Float.class, new NumberDeserializer())
                                    .registerTypeAdapter(Double.class, new NumberDeserializer())
                                    .create();
    }
    
    @Override
    public <T extends Dto> void fromString(String content, DtoContainer<T> result) {
        fromString(new StringReader(content), result);
    }

    @Override
    public <T extends Dto> String toString(DtoContainer<T> dto) {
        StringWriter writer = new StringWriter();
        toString(dto, writer);
        return writer.toString();
    }
    
    @Override
    @SuppressWarnings({ "unchecked", "rawtypes" })
    public <T extends Dto> void fromString(Reader reader, DtoContainer<T> result) {
        try {
            Class<T> dtoType = result.getDtoType();
            if (!registeredTypes.containsKey(dtoType)) {
                serializer.registerType(dtoType);
                registeredTypes.put(dtoType, VOID);
            }

            DtoContainer container = importer.fromJson(reader, DtoContainer.class);
            if (container == null) {
                result.clear();
            } else if(result.isMultiValued() && !container.isMultiValued()) {
                // strictly this should be an error but we can deal with this.
                if(container.getSingle() != null)
                    result.set(Arrays.asList((T) container.getSingle()));
            } else if (result.isMultiValued()) {
                // container is also multiple valued.
                result.set(container.get());
            } else if(container.isMultiValued()) {
                // result is single valued
                // if container has 1 or 0 elements we can still make it work.
                if(container.get().isEmpty()) {
                    result.clear();
                } else if (container.get().size() == 1) {
                    result.set((T)container.get().get(0));
                } else {
                    throw new ParseException("Expected only one object, recived array.");
                }
            } else {
                result.set((T)container.getSingle());
            }
        } catch (ParseException e) {
            // we allow the try body to throw parse Exceptions
            throw e;
        } catch(Exception e) {
            // Any other exception is treated as the root cause of a parse failure. 
            throw new ParseException(e);
        }
    }
    
    @Override
    public <T extends Dto> void toString(DtoContainer<T> dto, Writer writer) {
        try {
            if (!registeredTypes.containsKey(dto.getDtoType())) {
                serializer.registerType(dto.getDtoType());
                registeredTypes.put(dto.getDtoType(), VOID);
            }

            exporter.toJson(dto, writer);
        } catch(Exception e) {
            throw new ParseException(e);
        }
    }
    
    private String getDtoSingularName(Class<? extends Dto> type, Named name) {
        if (name != null)
            return name.singular();
        
        // if the first letter is capitalized, make it lower case.
        String derived = type.getSimpleName();
        String suffix = "Dto";
        derived = Character.toLowerCase(derived.charAt(0)) + derived.substring(1);
        
        // if the name ends with suffix, remove the suffix.
        if (derived.endsWith(suffix))
            derived = derived.substring(0, derived.length() - suffix.length());
        return derived;
    }
    
    private String getDtoPluralName(String singular, Named name) {
        if (name != null)
            return name.plural();
        return singular + "s"; // Dto default names always assumed to be singular
    }
    
    private class CharacterDeserializer implements JsonDeserializer<Character> {
        @Override
        public Character deserialize(JsonElement json, Type typeOfT, JsonDeserializationContext context) throws JsonParseException {
            if (!json.isJsonPrimitive())
                throw new JsonParseException("Expected character primitive");
            JsonPrimitive prim = json.getAsJsonPrimitive();
            String str = prim.getAsString();
            if (str.length() == 0)
                return Character.valueOf('\0');
            else if (str.length() == 1)
                return Character.valueOf(str.charAt(0));
            else
                throw new JsonParseException("Expected a single character");
        }
    }
    
    private class NumberDeserializer implements JsonDeserializer<Number> {
        @Override
        public Number deserialize(JsonElement json, Type typeOfT, JsonDeserializationContext context) throws JsonParseException {
            if (!json.isJsonPrimitive())
                throw new JsonParseException("Expected primitive of type " + typeOfT);

            String text = json.getAsString();
            if (typeOfT.equals(Double.class))
                return Double.parseDouble(text);
            else if (typeOfT.equals(Float.class))
                return Float.parseFloat(text);
            else if (typeOfT.equals(Long.class))
                return Long.parseLong(text);
            else if (typeOfT.equals(Integer.class))
                return Integer.parseInt(text);
            else if (typeOfT.equals(Short.class))
                return Short.parseShort(text);
            else if (typeOfT.equals(Byte.class))
                return Byte.parseByte(text);

            return null;
        }
    }
    
    private class DtoContainerSerializer implements JsonSerializer<DtoContainer<?>>, JsonDeserializer<DtoContainer<?>> {
        private final ConcurrentMap<String, Class<? extends Dto>> singleDtoMap = new ConcurrentHashMap<String, Class<? extends Dto>>();
        private final ConcurrentMap<String, Class<? extends Dto>> pluralDtoMap = new ConcurrentHashMap<String, Class<? extends Dto>>();
        
        public void registerType(Class<? extends Dto> type) {
            Named typeName = type.getAnnotation(Named.class);
            String single = getDtoSingularName(type, typeName);
            String plural = getDtoPluralName(single, typeName);
            
            // in the event that registerType is called simultaneously with
            // the same type, we're still okay since getDtoXName() will return
            // the same string values if the class type is unchanged
            singleDtoMap.put(single, type);
            pluralDtoMap.put(plural, type);
        }
        
        @Override
        @SuppressWarnings({ "unchecked", "rawtypes" })
        public DtoContainer<?> deserialize(JsonElement json, Type typeOfT, JsonDeserializationContext context) throws JsonParseException {
            if (!json.isJsonObject())
                throw new JsonParseException("Expected top-level json to be an object");
            
            JsonObject root = json.getAsJsonObject();
            if (root.entrySet().size() != 1)
                throw new JsonParseException("Expected top-level json to have a single property");

            Entry<String, JsonElement> child = root.entrySet().iterator().next();
            if (child.getValue().isJsonArray()) {
                // assuming we're dealing with multiples, expect plural name
                JsonArray array = child.getValue().getAsJsonArray();
                Class<? extends Dto> type = pluralDtoMap.get(child.getKey());
                if (type == null)
                    throw new JsonParseException("Unexpected property name: " + child.getKey());
                
                List<Dto> dtos = new ArrayList<Dto>(array.size());
                for (JsonElement el: array)
                    dtos.add((Dto) context.deserialize(el, type));
                
                return new DtoContainer(type, dtos);
            } else if (child.getValue().isJsonObject()) {
                // assume we're dealing with a single object
                Class<? extends Dto> type = singleDtoMap.get(child.getKey());
                if (type == null)
                    throw new JsonParseException("Unexpected property name: " + child.getKey());
                
                Dto wrapped = context.deserialize(child.getValue(), type);
                return new DtoContainer(type, wrapped);
            }
            
            throw new JsonParseException("Expected top-level json to have a single property that's an object or array");
        }
        
        @Override
        public JsonElement serialize(DtoContainer<?> src, Type typeOfSrc, JsonSerializationContext context) {
            Class<? extends Dto> type = src.getDtoType();
            Named typeName = type.getAnnotation(Named.class);
            
            String single = getDtoSingularName(type, typeName);
            JsonObject wrapper = new JsonObject();
            
            if(src.isMultiValued()) {
                String plural = getDtoPluralName(single, typeName);
                
                JsonArray array = new JsonArray();
                for (Dto d: src.get())
                    array.add(context.serialize(d));
                
                wrapper.add(plural, array);
            } else {
                wrapper.add(single, context.serialize(src.getSingle()));
            }
            return wrapper;
        }
    }
    
    /*
     * Internal class that handles excluding classes that aren't Dtos or boxed
     * primitives, or arrays of those types. Also excludes fields not properly
     * marked as Importable/Exportable
     */
    private class DtoExclusionStrategy implements ExclusionStrategy {
        private final boolean forInput;
        
        public DtoExclusionStrategy(boolean input) {
            forInput = input;
        }
        
        @Override
        public boolean shouldSkipClass(Class<?> clazz) {
            if (clazz.equals(DtoContainer.class)) // special case for root instances
                return false;
            
            if (clazz.isArray())
                clazz = clazz.getComponentType();
            if (Dto.class.isAssignableFrom(clazz) && !Modifier.isAbstract(clazz.getModifiers()))
                return false;
            if (Number.class.isAssignableFrom(clazz))
                return false;
            if (String.class.isAssignableFrom(clazz))
                return false;
            if (Boolean.class.isAssignableFrom(clazz))
                return false;
            if (Character.class.isAssignableFrom(clazz))
                return false;
            return true;
        }

        @Override
        public boolean shouldSkipField(FieldAttributes f) {
            // we add the DtoContainer because we want to allow root DtoContainers to be serialized,
            // but not ones that are used as fields
            if (!f.hasModifier(Modifier.PUBLIC) || f.getDeclaredClass().isAssignableFrom(DtoContainer.class))
                return true;
            
            if (!ignoreExclude) {
                Exclude exclude = f.getAnnotation(Exclude.class);
                if (exclude != null) {
                    ExcludeType[] types = exclude.value();
                    ExcludeType skipIfPresent = (forInput ? ExcludeType.IMPORT : ExcludeType.EXPORT);
                    for (int i = 0; i < types.length; i++) {
                        if (skipIfPresent.equals(types[i]))
                            return true;
                    }
                }
            }

            return false;
        }
    }
    
    /*
     * Internal class that implements a NamingStrategy that respects the rules layed
     * out by the Named annotation.
     */
    private static class DtoNamingStrategy implements FieldNamingStrategy {
        
        @Override
        public String translateName(Field f) {
            Named n = f.getAnnotation(Named.class);
            if (n != null) {
                return (f.getType().isArray() ? n.plural() : n.singular());
            } else
                return f.getName();
        }
    }
}
